#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES 1D Fixed Bed model.

"""

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Param,
    Reals,
    value,
    TransformationFactory,
    Constraint,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.util.subsystems import TemporarySubsystemManager
from pyomo.common.collections import ComponentSet
from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
    solve_strongly_connected_components,
)

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    DistributedVars,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_reaction_parameter_block,
)
from idaes.core.util.exceptions import (
    ConfigurationError,
    BurntToast,
    InitializationError,
)
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.dyn_utils import get_index_set_except
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.constants import Constants as constants
from idaes.core.util.math import smooth_abs
import idaes.logger as idaeslog
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

__author__ = "Chinedu Okoli"

# Set up logger
_log = idaeslog.getLogger(__name__)

# Assumptions:
# Only gas and solid phases, which are explicitly named as such
# Perfect mixing in gas phase
# Axially varying gas and solid phase
# Reaction/adsorption on the solid is the rate limiting step


@declare_process_block_class("FixedBed1D")
class FixedBed1DData(UnitModelBlockData):
    """
    1D Fixed Bed Reactor Model Class
    """

    # Create template for unit level config arguments
    CONFIG = UnitModelBlockData.CONFIG()

    # Unit level config arguments
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=10,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "length_domain_set",
        ConfigValue(
            default=[0.0, 1.0],
            domain=list,
            description="Number of finite elements length domain",
            doc="""length_domain_set - (optional) list of point to use to
initialize a new ContinuousSet if length_domain is not
provided (default = [0.0, 1.0])""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default="dae.finite_difference",
            description="Method to use for DAE transformation",
            doc="""Method to use to transform domain. Must be a method recognized
by the Pyomo TransformationFactory,
**default** - "dae.finite_difference".
**Valid values:** {
**"dae.finite_difference"** - Use a finite difference transformation method,
**"dae.collocation"** - use a collocation transformation method}""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=None,
            domain=In([None, "BACKWARD", "FORWARD", "LAGRANGE-RADAU"]),
            description="Scheme to use for DAE transformation",
            doc="""Scheme to use when transforming domain. See Pyomo
documentation for supported schemes,
**default** - None.
**Valid values:** {
**None** - defaults to "BACKWARD" for finite difference transformation method,
and to "LAGRANGE-RADAU" for collocation transformation method,
**"BACKWARD"** - Use a finite difference transformation method,
**"FORWARD""** - use a finite difference transformation method,
**"LAGRANGE-RADAU""** - use a collocation transformation method}""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=3,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "flow_type",
        ConfigValue(
            default="forward_flow",
            domain=In(["forward_flow", "reverse_flow"]),
            description="Flow configuration of Fixed Bed",
            doc="""Flow configuration of Fixed Bed
**default** - "forward_flow".
**Valid values:** {
**"forward_flow"** - gas flows from 0 to 1,
**"reverse_flow"** -  gas flows from 1 to 0.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentTotal,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentTotal.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "pressure_drop_type",
        ConfigValue(
            default="ergun_correlation",
            domain=In(["ergun_correlation", "simple_correlation"]),
            description="Construction flag for type of pressure drop",
            doc="""Indicates what type of pressure drop correlation should be used,
**default** - "ergun_correlation".
**Valid values:** {
**"ergun_correlation"** - Use the Ergun equation,
**"simple_correlation"** - Use a simplified pressure drop correlation.}""",
        ),
    )

    # Create template for phase specific config arguments
    _PhaseTemplate = UnitModelBlockData.CONFIG()
    _PhaseTemplate.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    _PhaseTemplate.declare(
        "property_package_args",
        ConfigValue(
            default={},
            domain=dict,
            description="Arguments for constructing gas property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    _PhaseTemplate.declare(
        "reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    _PhaseTemplate.declare(
        "reaction_package_args",
        ConfigBlock(
            implicit=True,
            implicit_domain=ConfigBlock,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}""",
        ),
    )
    _PhaseTemplate.declare(
        "has_equilibrium_reactions",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Equilibrium reaction construction flag",
            doc="""Indicates whether terms for equilibrium controlled reactions
should be constructed,
**default** - True.
**Valid values:** {
**True** - include equilibrium reaction terms,
**False** - exclude equilibrium reaction terms.}""",
        ),
    )

    # Create individual config blocks for gas and solid sides
    CONFIG.declare("gas_phase_config", _PhaseTemplate(doc="gas phase config arguments"))
    CONFIG.declare(
        "solid_phase_config", _PhaseTemplate(doc="solid phase config arguments")
    )

    # =========================================================================
    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to build default attributes
        super().build()

        # local aliases used to shorten object names
        solid_phase = self.config.solid_phase_config
        gas_phase = self.config.gas_phase_config

        # Get units meta data from property packages
        units_meta_solid = solid_phase.property_package.get_metadata().get_derived_units

        # Set flow direction for the gas control volume
        # Gas flows from 0 to 1
        if self.config.flow_type == "forward_flow":
            set_direction_gas = FlowDirection.forward
        # Gas flows from 1 to 0
        if self.config.flow_type == "reverse_flow":
            set_direction_gas = FlowDirection.backward

        # Consistency check for flow direction, transformation method and
        # transformation scheme
        if (
            self.config.flow_type == "forward_flow"
            and self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme is None
        ):
            self.config.transformation_scheme = "BACKWARD"
        elif (
            self.config.flow_type == "reverse_flow"
            and self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme is None
        ):
            self.config.transformation_scheme = "FORWARD"
        elif (
            self.config.flow_type == "forward_flow"
            and self.config.transformation_method == "dae.collocation"
            and self.config.transformation_scheme is None
        ):
            self.config.transformation_scheme = "LAGRANGE-RADAU"
        elif (
            self.config.flow_type == "reverse_flow"
            and self.config.transformation_method == "dae.collocation"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_method argument."
                "Must be "
                "dae.finite_difference "
                "if "
                "flow_type is"
                " "
                "reverse_flow"
                ".".format(self.name)
            )
        elif (
            self.config.flow_type == "forward_flow"
            and self.config.transformation_scheme == "FORWARD"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument. "
                "Must be "
                "BACKWARD "
                "if flow_type is"
                " "
                "forward_flow"
                ".".format(self.name)
            )
        elif (
            self.config.flow_type == "reverse_flow"
            and self.config.transformation_scheme == "BACKWARD"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument."
                "Must be "
                "FORWARD "
                "if "
                "flow_type is"
                " "
                "reverse_flow"
                ".".format(self.name)
            )
        elif (
            self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme != "BACKWARD"
            and self.config.transformation_scheme != "FORWARD"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument. "
                "Must be "
                "BACKWARD"
                " or "
                "FORWARD"
                " "
                "if transformation_method is"
                " "
                "dae.finite_difference"
                ".".format(self.name)
            )
        elif (
            self.config.transformation_method == "dae.collocation"
            and self.config.transformation_scheme != "LAGRANGE-RADAU"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument."
                "Must be "
                "LAGRANGE-RADAU"
                " if "
                "transformation_method is"
                " "
                "dae.collocation"
                ".".format(self.name)
            )

        # Set arguments for gas sides if homoogeneous reaction block
        if gas_phase.reaction_package is not None:
            has_rate_reaction_gas_phase = True
        else:
            has_rate_reaction_gas_phase = False

        # Set arguments for gas and solid sides if heterogeneous reaction block
        if solid_phase.reaction_package is not None:
            has_mass_transfer_gas_phase = True
        else:
            has_mass_transfer_gas_phase = False

        # Set heat transfer terms
        if self.config.energy_balance_type != EnergyBalanceType.none:
            has_heat_transfer = True
        else:
            has_heat_transfer = False

        # Set heat of reaction terms
        if (
            self.config.energy_balance_type != EnergyBalanceType.none
            and gas_phase.reaction_package is not None
        ):
            has_heat_of_reaction_gas_phase = True
        else:
            has_heat_of_reaction_gas_phase = False

        # Create a unit model length domain
        self.length_domain = ContinuousSet(
            bounds=(0.0, 1.0),
            initialize=self.config.length_domain_set,
            doc="Normalized length domain",
        )

        self.bed_height = Var(
            domain=Reals,
            initialize=1,
            doc="Bed length",
            units=units_meta_solid("length"),
        )

        # =========================================================================
        # Build Control volume 1D for gas phase and populate gas control volume

        self.gas_phase = ControlVolume1DBlock(
            transformation_method=self.config.transformation_method,
            transformation_scheme=self.config.transformation_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
            dynamic=True,  # Fixed beds must be dynamic
            has_holdup=True,  # holdup must be True for fixed beds
            area_definition=DistributedVars.variant,
            property_package=gas_phase.property_package,
            property_package_args=gas_phase.property_package_args,
            reaction_package=gas_phase.reaction_package,
            reaction_package_args=gas_phase.reaction_package_args,
        )

        self.gas_phase.add_geometry(
            length_domain=self.length_domain,
            length_domain_set=self.config.length_domain_set,
            length_var=self.bed_height,
            flow_direction=set_direction_gas,
        )

        self.gas_phase.add_state_blocks(
            information_flow=set_direction_gas, has_phase_equilibrium=False
        )

        if gas_phase.reaction_package is not None:
            self.gas_phase.add_reaction_blocks(
                has_equilibrium=gas_phase.has_equilibrium_reactions
            )

        self.gas_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
            has_mass_transfer=has_mass_transfer_gas_phase,
            has_rate_reactions=has_rate_reaction_gas_phase,
        )

        self.gas_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=has_heat_transfer,
            has_heat_of_reaction=has_heat_of_reaction_gas_phase,
        )

        self.gas_phase.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # =========================================================================
        # Build Control volume 1D for solid phase and populate solid control volume

        # Build Solid Phase StateBlock
        # As there is no solid flow, there is not need for a control volume as a
        # set of indexed state blocks will be sufficient.
        # Defined state is set to True so that the "sum(mass_frac)=1" eqn in
        # the solid state block is deactivated. This is done here as there is
        # currently no way to deactivate the constraint at the initial time
        # for batch systems (i.e. no inlet or outlet ports).
        # The "sum(mass_frac)=1 for all t neq 0" eqn is written in the
        # unit model instead
        self.solid_properties = solid_phase.property_package.state_block_class(
            self.flowsheet().time,
            self.length_domain,
            parameters=solid_phase.property_package,
            defined_state=True,
        )

        if solid_phase.reaction_package is not None:
            # TODO - a generalization of the heterogeneous reaction block
            # The heterogeneous reaction block does not use the
            # add_reaction_blocks in control volumes as control volumes are
            # currently setup to handle only homogeneous reaction properties.
            # Thus appending the heterogeneous reaction block to the
            # solid state block is currently hard coded here.

            tmp_dict = dict(**solid_phase.reaction_package_args)
            tmp_dict["gas_state_block"] = self.gas_phase.properties
            tmp_dict["solid_state_block"] = self.solid_properties
            tmp_dict["parameters"] = solid_phase.reaction_package
            self.solid_reactions = solid_phase.reaction_package.reaction_block_class(
                self.flowsheet().time,
                self.length_domain,
                doc="Reaction properties in control volume",
                **tmp_dict,
            )

        # =========================================================================
        # Add Ports for gas side
        self.add_inlet_port(name="gas_inlet", block=self.gas_phase)
        self.add_outlet_port(name="gas_outlet", block=self.gas_phase)

        # =========================================================================
        # Add performance equation method
        self._apply_transformation()
        self._make_performance()

    # =========================================================================
    def _apply_transformation(self):
        """
        Method to apply DAE transformation to the Control Volume length domain.
        Transformation applied will be based on the Control Volume
        configuration arguments.
        """
        if self.config.finite_elements is None:
            raise ConfigurationError(
                "{} was not provided a value for the finite_elements"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )

        if self.config.transformation_method == "dae.finite_difference":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                wrt=self.length_domain,
                nfe=self.config.finite_elements,
                scheme=self.config.transformation_scheme,
            )
        elif self.config.transformation_method == "dae.collocation":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                wrt=self.length_domain,
                nfe=self.config.finite_elements,
                ncp=self.config.collocation_points,
                scheme=self.config.transformation_scheme,
            )

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        # local aliases used to shorten object names
        gas_phase = self.config.gas_phase_config
        solid_phase = self.config.solid_phase_config

        # Get units meta data from property packages
        units_meta_gas = gas_phase.property_package.get_metadata().get_derived_units
        units_meta_solid = solid_phase.property_package.get_metadata().get_derived_units

        # Declare Mutable Parameters
        self.eps = Param(
            mutable=True,
            default=1e-8,
            doc="Smoothing Factor for Smooth IF Statements",
            units=pyunits.dimensionless,
        )

        # Unit Model variables
        self.bed_diameter = Var(
            initialize=1, doc="Reactor diameter", units=units_meta_solid("length")
        )
        self.bed_area = Var(
            domain=Reals,
            initialize=1,
            doc="Reactor cross-sectional area",
            units=units_meta_solid("area"),
        )

        # Phase specific variables
        self.velocity_superficial_gas = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=0.05,
            doc="Gas superficial velocity",
            units=units_meta_gas("velocity"),
        )

        # Dimensionless numbers, mass and heat transfer coefficients
        self.Re_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Particle Reynolds number",
            units=pyunits.dimensionless,
        )

        self.Pr_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Prandtl number of gas in bed",
            units=pyunits.dimensionless,
        )

        self.Nu_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Particle Nusselt number",
            units=pyunits.dimensionless,
        )
        self.gas_solid_htc = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Gas-solid heat transfer coefficient",
            units=units_meta_gas("heat_transfer_coefficient"),
        )

        # Solid phase variables
        self.solid_phase_area = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1,
            doc="Solid phase area",
            units=units_meta_solid("area"),
        )

        self.solid_material_holdup = Var(
            self.flowsheet().time,
            self.length_domain,
            solid_phase.property_package.component_list,
            domain=Reals,
            initialize=1.0,
            doc="Solid material holdup per unit length",
            units=units_meta_solid("mass") / units_meta_solid("length"),
        )

        self.solid_material_accumulation = DerivativeVar(
            self.solid_material_holdup,
            wrt=self.flowsheet().config.time,
            doc="Solid material accumulation",
            units=units_meta_solid("mass")
            / units_meta_solid("length")
            / units_meta_solid("time"),
        )

        if self.config.energy_balance_type != EnergyBalanceType.none:
            # Solid phase - gas to solid heat transfer
            self.solid_phase_heat = Var(
                self.flowsheet().time,
                self.length_domain,
                domain=Reals,
                initialize=0.0,
                doc="Heat transferred per unit length",
                units=units_meta_solid("power") / units_meta_solid("length"),
            )

            self.solid_energy_holdup = Var(
                self.flowsheet().time,
                self.length_domain,
                initialize=1,
                doc="Solid phase energy holdup",
                units=units_meta_solid("energy") / units_meta_solid("length"),
            )

            self.solid_energy_accumulation = DerivativeVar(
                self.solid_energy_holdup,
                initialize=0,
                wrt=self.flowsheet().config.time,
                doc="Solid energy accumulation",
                units=units_meta_solid("power") / units_meta_solid("length"),
            )

        # =========================================================================
        # Add performance equations

        # ---------------------------------------------------------------------
        # TODO - Deactivate/delete sum_component_eqns of initial conditions for
        #  all space in gas_phase as currently no way to tell CV1D to not build
        #  sum_eqns for initial conditions
        if self.gas_phase.config.dynamic:
            t0 = self.flowsheet().time.first()
            for x in self.length_domain:
                blk = self.gas_phase.properties[t0, x]
                if hasattr(blk, "sum_component_eqn"):
                    blk.sum_component_eqn.deactivate()

        # ---------------------------------------------------------------------
        # Geometry constraints

        # Bed area
        @self.Constraint(doc="Bed area")
        def bed_area_eqn(b):
            return b.bed_area == (constants.pi * (0.5 * b.bed_diameter) ** 2)

        # Area of gas side, and solid side
        @self.Constraint(self.flowsheet().time, self.length_domain, doc="Gas side area")
        def gas_phase_area_constraint(b, t, x):
            return (
                b.gas_phase.area[t, x]
                == pyunits.convert(b.bed_area, to_units=units_meta_gas("area"))
                * b.solid_properties[t, x].params.voidage
            )

        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Solid side area"
        )
        def solid_phase_area_constraint(b, t, x):
            return b.solid_phase_area[t, x] == b.bed_area * (
                1 - b.solid_properties[t, x].params.voidage
            )

        # ---------------------------------------------------------------------
        # Hydrodynamic constraints

        # Gas superficial velocity
        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Gas superficial velocity"
        )
        def gas_super_vel(b, t, x):
            return (
                b.velocity_superficial_gas[t, x]
                * pyunits.convert(b.bed_area, to_units=units_meta_gas("area"))
                * b.gas_phase.properties[t, x].dens_mol
                == b.gas_phase.properties[t, x].flow_mol
            )

        # Gas side pressure drop calculation
        if (
            self.config.has_pressure_change
            and self.config.pressure_drop_type == "simple_correlation"
        ):
            # Simplified pressure drop
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas side pressure drop calculation -" "simplified pressure drop",
            )
            def gas_phase_config_pressure_drop(b, t, x):
                #  0.2/s is a unitted constant in the correlation
                return pyunits.convert(
                    b.gas_phase.deltaP[t, x],
                    to_units=units_meta_solid("pressure") / units_meta_solid("length"),
                ) == -(0.2 / pyunits.s) * (
                    b.velocity_superficial_gas[t, x]
                    * (
                        pyunits.convert(
                            b.solid_properties[t, x].dens_mass_particle,
                            to_units=units_meta_solid("density_mass"),
                        )
                        - b.gas_phase.properties[t, x].dens_mass
                    )
                )

        elif (
            self.config.has_pressure_change
            and self.config.pressure_drop_type == "ergun_correlation"
        ):
            # Ergun equation
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas side pressure drop calculation -" "Ergun equation",
            )
            def gas_phase_config_pressure_drop(b, t, x):
                return -pyunits.convert(
                    b.gas_phase.deltaP[t, x],
                    to_units=units_meta_solid("pressure") / units_meta_solid("length"),
                ) == (
                    (150 * pyunits.dimensionless)
                    * (1 - b.solid_properties[t, x].params.voidage) ** 2
                    * b.gas_phase.properties[t, x].visc_d
                    * b.velocity_superficial_gas[t, x]
                    / (
                        pyunits.convert(
                            b.solid_properties[t, x].params.particle_dia,
                            to_units=units_meta_solid("length"),
                        )
                        ** 2
                        * b.solid_properties[t, x].params.voidage ** 3
                    )
                ) + (
                    (1.75 * pyunits.dimensionless)
                    * b.gas_phase.properties[t, x].dens_mass
                    * (1 - b.solid_properties[t, x].params.voidage)
                    * b.velocity_superficial_gas[t, x] ** 2
                    / (
                        pyunits.convert(
                            b.solid_properties[t, x].params.particle_dia,
                            to_units=units_meta_solid("length"),
                        )
                        * b.solid_properties[t, x].params.voidage ** 3
                    )
                )

        elif self.config.has_pressure_change is False:
            pass

        else:
            raise BurntToast(
                "{} encountered unrecognized argument for "
                "the pressure drop correlation. Please contact the IDAES"
                " developers with this bug.".format(self.name)
            )
        # ---------------------------------------------------------------------
        # Reaction constraints

        # Build homogeneous reaction constraints
        if gas_phase.reaction_package is not None:
            # Gas side rate reaction extent
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                gas_phase.reaction_package.rate_reaction_idx,
                doc="Gas side rate reaction extent",
            )
            def gas_phase_config_rxn_ext(b, t, x, r):
                return b.gas_phase.rate_reaction_extent[t, x, r] == (
                    b.gas_phase.reactions[t, x].reaction_rate[r]
                    * b.gas_phase.area[t, x]
                )

        # Build hetereogeneous reaction constraints
        if solid_phase.reaction_package is not None:
            # Gas side heterogeneous rate reaction generation
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                gas_phase.property_package.phase_list,
                gas_phase.property_package.component_list,
                doc="Gas side heterogeneous" "rate reaction generation",
            )
            def gas_comp_hetero_rxn(b, t, x, p, j):
                return b.gas_phase.mass_transfer_term[t, x, p, j] == (
                    sum(
                        b.solid_reactions[t, x].rate_reaction_stoichiometry[r, p, j]
                        * b.solid_reactions[t, x].reaction_rate[r]
                        for r in (solid_phase.reaction_package.rate_reaction_idx)
                    )
                    * b.solid_phase_area[t, x]
                )

        # ---------------------------------------------------------------------
        # Solid phase component balance

        # material holdup constraint
        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            solid_phase.property_package.component_list,
            doc="Solid phase material holdup constraints",
        )
        def solid_material_holdup_calculation(b, t, x, j):
            return b.solid_material_holdup[t, x, j] == (
                b.solid_phase_area[t, x]
                * b.solid_properties[t, x].get_material_density_terms("Sol", j)
            )

        # Add component balances
        @self.Constraint(
            self.flowsheet().time,
            self.length_domain,
            solid_phase.property_package.component_list,
            doc="Material balances",
        )
        def solid_material_balances(b, t, x, j):
            if solid_phase.reaction_package is not None:
                return b.solid_material_accumulation[t, x, j] == (
                    b.solid_phase_area[t, x]
                    * b.solid_properties[t, x].params.mw_comp[j]
                    * sum(
                        b.solid_reactions[t, x].reaction_rate[r]
                        * solid_phase.reaction_package.rate_reaction_stoichiometry[
                            r, "Sol", j
                        ]
                        for r in solid_phase.reaction_package.rate_reaction_idx
                    )
                )
            else:
                return b.solid_material_accumulation[t, x, j] == 0

        # Sum of mass fractions at all time equals 1
        @self.Constraint(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Sum of mass fractions at all time",
        )
        def solid_sum_component_eqn(b, t, x):
            if t == b.flowsheet().config.time.first():
                return Constraint.Skip
            else:
                return 1 == sum(
                    b.solid_properties[t, x].mass_frac_comp[j]
                    for j in b.solid_properties[t, x].params.component_list
                )

        # ---------------------------------------------------------------------
        if self.config.energy_balance_type != EnergyBalanceType.none:
            # Solid phase - gas to solid heat transfer
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Solid phase - gas to solid heat transfer",
            )
            def solid_phase_heat_transfer(b, t, x):
                return pyunits.convert(
                    b.solid_phase_heat[t, x],
                    to_units=units_meta_gas("power") / units_meta_gas("length"),
                ) * pyunits.convert(
                    b.solid_properties[t, x].params.particle_dia,
                    to_units=units_meta_gas("length"),
                ) == 6 * b.gas_solid_htc[
                    t, x
                ] * (
                    b.gas_phase.properties[t, x].temperature
                    - pyunits.convert(
                        b.solid_properties[t, x].temperature,
                        to_units=units_meta_gas("temperature"),
                    )
                ) * pyunits.convert(
                    b.solid_phase_area[t, x], to_units=units_meta_gas("area")
                )

            # Solid phase energy balance
            # Accumulation equal to heat transfer
            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Solid phase energy holdup constraints",
            )
            def solid_energy_holdup_calculation(b, t, x):
                return b.solid_energy_holdup[t, x] == (
                    b.solid_phase_area[t, x]
                    * pyunits.convert(
                        b.solid_properties[t, x].get_energy_density_terms("Sol"),
                        to_units=units_meta_solid("energy")
                        / units_meta_solid("volume"),
                    )
                )

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Solid phase energy balances",
            )
            def solid_enthalpy_balances(b, t, x):
                if solid_phase.reaction_package is not None:
                    return b.solid_energy_accumulation[t, x] / units_meta_solid(
                        "time"
                    ) == b.solid_phase_heat[t, x] + -sum(
                        b.solid_reactions[t, x].reaction_rate[r]
                        * b.solid_phase_area[t, x]
                        * pyunits.convert(
                            b.solid_reactions[t, x].dh_rxn[r],
                            to_units=units_meta_solid("energy_mole"),
                        )
                        for r in solid_phase.reaction_package.rate_reaction_idx
                    )
                else:
                    return b.solid_energy_accumulation[t, x] == b.solid_phase_heat[t, x]

            # Dimensionless numbers, mass and heat transfer coefficients
            # Particle Reynolds number
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Particle Reynolds number",
            )
            def reynolds_number_particle(b, t, x):
                return (
                    b.Re_particle[t, x] * b.gas_phase.properties[t, x].visc_d
                    == b.velocity_superficial_gas[t, x]
                    * pyunits.convert(
                        b.solid_properties[t, x].params.particle_dia,
                        to_units=units_meta_gas("length"),
                    )
                    * b.gas_phase.properties[t, x].dens_mass
                )

            # Prandtl number
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Prandtl number of gas in bed",
            )
            def prandtl_number(b, t, x):
                return (
                    b.Pr_particle[t, x] * b.gas_phase.properties[t, x].therm_cond
                    == b.gas_phase.properties[t, x].cp_mass
                    * b.gas_phase.properties[t, x].visc_d
                )

            # Particle Nusselt number
            @self.Constraint(
                self.flowsheet().time, self.length_domain, doc="Particle Nusselt number"
            )
            def nusselt_number_particle(b, t, x):
                return (
                    b.Nu_particle[t, x] ** 3
                    == (
                        2.0 + 1.1 * (smooth_abs(b.Re_particle[t, x], b.eps) ** 0.6) ** 3
                    )
                    * b.Pr_particle[t, x]
                )

            # Gas-solid heat transfer coefficient
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas-solid heat transfer coefficient",
            )
            def gas_solid_htc_eqn(b, t, x):
                return (
                    b.gas_solid_htc[t, x]
                    * pyunits.convert(
                        b.solid_properties[t, x].params.particle_dia,
                        to_units=units_meta_gas("length"),
                    )
                    == b.Nu_particle[t, x] * b.gas_phase.properties[t, x].therm_cond
                )

            # Gas phase - gas to solid heat transfer
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas phase - gas to solid heat transfer",
            )
            def gas_phase_heat_transfer(b, t, x):
                return b.gas_phase.heat[t, x] * pyunits.convert(
                    b.solid_properties[t, x].params.particle_dia,
                    to_units=units_meta_gas("length"),
                ) == -6 * b.gas_solid_htc[t, x] * (
                    b.gas_phase.properties[t, x].temperature
                    - pyunits.convert(
                        b.solid_properties[t, x].temperature,
                        to_units=units_meta_gas("temperature"),
                    )
                ) * pyunits.convert(
                    b.solid_phase_area[t, x], to_units=units_meta_gas("area")
                )

        elif self.config.energy_balance_type == EnergyBalanceType.none:
            # if self.config.energy_balance_type == EnergyBalanceType.none:
            # If energy balance is none fix gas and solid temperatures to
            # initial solid temperature
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Isothermal gas phase constraint",
            )
            def isothermal_gas_phase(b, t, x):
                # Skip constraint at the initial and boundary points
                t0 = b.flowsheet().config.time.first()
                if b.config.flow_type == "forward_flow":
                    x_inlet = b.length_domain.first()
                else:
                    x_inlet = b.length_domain.last()
                if x == x_inlet or t == b.flowsheet().config.time.first():
                    return Constraint.Skip
                else:
                    return (
                        b.gas_phase.properties[t, x].temperature
                        == b.solid_properties[t0, x].temperature
                    )

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="Isothermal solid phase constraint",
            )
            def isothermal_solid_phase(b, t, x):
                # Skip constraint at the initial point
                t0 = b.flowsheet().config.time.first()
                if t == b.flowsheet().config.time.first():
                    return Constraint.Skip
                else:
                    return (
                        b.solid_properties[t, x].temperature
                        == b.solid_properties[t0, x].temperature
                    )

        # Set default scaling values for some CV1D vars (to avoid warnings)
        if hasattr(self.gas_phase, "heat"):
            for v in self.gas_phase.heat.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-6)
        if hasattr(self.gas_phase, "area"):
            for v in self.gas_phase.area.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

    # =========================================================================
    # Model initialization routine

    def initialize_build(
        blk,
        gas_phase_state_args=None,
        solid_phase_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for 1DFixedBed unit.

        Keyword Arguments:
            gas_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            solid_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # local aliases used to shorten object names
        gas_phase = blk.config.gas_phase_config
        solid_phase = blk.config.solid_phase_config

        # Keep all unit model geometry constraints, derivative_var constraints,
        # and property block constraints active. Additionally, in control
        # volumes - keep conservation linking constraints and
        # holdup calculation (for dynamic flowsheets) constraints active

        geometry_constraints_terms = [
            "bed_area_eqn",
            "solid_phase_area_constraint",
            "gas_phase_area_constraint",
        ]
        endswith_terms = (
            "_disc_eq",
            "linking_constraint",
            "linking_constraints",
            "_holdup_calculation",
        )
        startswith_terms = ("properties", "solid_properties")

        for c in blk.component_objects(Constraint, descend_into=True):
            if (
                not c.parent_block().local_name.startswith(startswith_terms)
                and not c.local_name.endswith(endswith_terms)
                and c.local_name not in geometry_constraints_terms
            ):
                c.deactivate()

        # ---------------------------------------------------------------------
        # Initialize thermophysical property constraints
        init_log.info("Initialize Thermophysical Properties")
        # Initialize gas_phase block
        gas_phase_flags = blk.gas_phase.properties.initialize(
            state_args=gas_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Initialize solid_phase properties block
        solid_phase_flags = blk.solid_properties.initialize(
            state_args=solid_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Initialize hydrodynamics (gas velocity)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                calculate_variable_from_constraint(
                    blk.velocity_superficial_gas[t, x], blk.gas_super_vel[t, x]
                )

        blk.gas_super_vel.activate()

        init_log.info("Initialize Hydrodynamics")
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)
        if check_optimal_termination(results):
            init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(results))
            )
        else:
            _log.warning("{} Initialization Step 2 Failed.".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize mass balance - no reaction and no pressure drop

        # Unfix material balance state variables (including particle porosity)
        # but keep other states fixed
        blk.gas_phase.properties.release_state(gas_phase_flags)
        blk.solid_properties.release_state(solid_phase_flags)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                blk.gas_phase.properties[t, x].pressure.fix()
                blk.gas_phase.properties[t, x].temperature.fix()
                blk.solid_properties[t, x].temperature.fix()

        if blk.config.has_holdup is True:
            # Fix initial conditions of flowrate
            t0 = blk.flowsheet().time.first()
            if blk.config.flow_type == "forward_flow":
                x_inlet = blk.length_domain.first()
            else:
                x_inlet = blk.length_domain.last()
            for x in blk.length_domain:
                if x != x_inlet:
                    blk.gas_phase.properties[t0, x].flow_mol.fix()
                    blk.gas_phase.properties[t0, x].sum_component_eqn.deactivate()

        blk.gas_phase.material_balances.activate()

        if gas_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                gas_rxn_gen = blk.gas_phase.rate_reaction_generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].fix(0.0))

        blk.solid_material_balances.activate()
        blk.solid_sum_component_eqn.activate()

        if solid_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for r in solid_phase.reaction_package.rate_reaction_idx:
                        blk.solid_reactions[t, x].reaction_rate[r].fix(0.0)

                # Gas side heterogeneous rate reaction generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (blk.gas_phase.mass_transfer_term[t, x, p, j].fix(0.0))

        init_log.info("Initialize Mass Balances")
        init_log.info_high(
            "initialize mass balances - no reactions and no pressure drop"
        )
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee, symbolic_solver_labels=True)
        if check_optimal_termination(results):
            init_log.info_high(
                "Initialization Step 3a {}.".format(idaeslog.condition(results))
            )
        else:
            _log.warning("{} Initialization Step 3a Failed.".format(blk.name))

        # Initialize mass balance - with reaction and no pressure drop
        if gas_phase.reaction_package is not None:
            # local aliases used to shorten object names
            gas_rxn_gen = blk.gas_phase.rate_reaction_generation
            gas_phase_stoichiometry_eqn = (
                blk.gas_phase.rate_reaction_stoichiometry_constraint
            )

            # Initialize reaction property package
            blk.gas_phase.reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    obj = blk.gas_phase.reactions[t, x]
                    for c in obj.component_objects(Constraint, descend_into=False):
                        c.activate()

            blk.gas_phase.reactions.initialize(
                outlvl=outlvl, optarg=optarg, solver=solver
            )

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for r in gas_phase.reaction_package.rate_reaction_idx:
                        calculate_variable_from_constraint(
                            blk.gas_phase.rate_reaction_extent[t, x, r],
                            blk.gas_phase_config_rxn_ext[t, x, r],
                        )
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].unfix())
                            if not (
                                (
                                    blk.gas_phase.config.transformation_scheme
                                    != "FORWARD"
                                    and x == blk.length_domain.first()
                                )
                                or (
                                    blk.gas_phase.config.transformation_scheme
                                    == "FORWARD"
                                    and x == blk.length_domain.last()
                                )
                            ):
                                calculate_variable_from_constraint(
                                    gas_rxn_gen[t, x, p, j],
                                    gas_phase_stoichiometry_eqn[t, x, p, j],
                                )

            gas_phase_stoichiometry_eqn.activate()
            blk.gas_phase_config_rxn_ext.activate()

        if solid_phase.reaction_package is not None:
            gas_mass_transfer_term = blk.gas_phase.mass_transfer_term

            # Initialize reaction property package
            blk.solid_reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    blk.solid_reactions[t, x].reaction_rate.unfix()
                    obj = blk.solid_reactions[t, x]
                    for c in obj.component_objects(Constraint, descend_into=False):
                        c.activate()

            blk.solid_reactions.initialize(outlvl=outlvl, optarg=optarg, solver=solver)

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_mass_transfer_term[t, x, p, j].unfix())
                            calculate_variable_from_constraint(
                                gas_mass_transfer_term[t, x, p, j],
                                blk.gas_comp_hetero_rxn[t, x, p, j],
                            )

            blk.gas_comp_hetero_rxn.activate()

        if (
            gas_phase.reaction_package is not None
            or solid_phase.reaction_package is not None
        ):
            init_log.info_high(
                "initialize mass balances - with reactions and no pressure drop"
            )
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee, symbolic_solver_labels=True)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 3b {}.".format(idaeslog.condition(results))
                )
            else:
                _log.warning("{} Initialization Step 3b Failed.".format(blk.name))

        # Initialize mass balance - with pressure drop
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                # Unfix all pressure variables except at the inlet
                if blk.gas_phase.properties[t, x].config.defined_state is False:
                    blk.gas_phase.properties[t, x].pressure.unfix()

        blk.gas_phase.pressure_balance.activate()

        if blk.config.has_pressure_change:
            blk.gas_phase_config_pressure_drop.activate()

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    calculate_variable_from_constraint(
                        blk.gas_phase.deltaP[t, x],
                        blk.gas_phase_config_pressure_drop[t, x],
                    )

            init_log.info_high(
                "initialize mass balances - with reactions and pressure drop"
            )
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee, symbolic_solver_labels=True)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 3c {}.".format(idaeslog.condition(results))
                )
            else:
                _log.warning("{} Initialization Step 3c Failed.".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize energy balance
        calc_var_kwds = {"eps": 5e-6}
        if blk.config.energy_balance_type != EnergyBalanceType.none:
            # Unfix temperatures
            if blk.config.flow_type == "forward_flow":
                x_inlet = blk.length_domain.first()
            else:
                x_inlet = blk.length_domain.last()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    if t != blk.flowsheet().config.time.first() and x != x_inlet:
                        # Unfix gas temperature variables except at the inlet
                        blk.gas_phase.properties[t, x].temperature.unfix()

            for t in blk.flowsheet().time:
                if t != blk.flowsheet().config.time.first():
                    for x in blk.length_domain:
                        # Unfix solid temperature variables except at the inlet
                        blk.solid_properties[t, x].temperature.unfix()

            # Initialize dimensionless numbers,
            # mass and heat transfer coefficients
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    # Initialize gas temperature to solid temperature values
                    calculate_variable_from_constraint(
                        blk.Re_particle[t, x],
                        blk.reynolds_number_particle[t, x],
                        **calc_var_kwds,
                    )
                    calculate_variable_from_constraint(
                        blk.Pr_particle[t, x], blk.prandtl_number[t, x], **calc_var_kwds
                    )
                    calculate_variable_from_constraint(
                        blk.Nu_particle[t, x],
                        blk.nusselt_number_particle[t, x],
                        **calc_var_kwds,
                    )
                    calculate_variable_from_constraint(
                        blk.gas_solid_htc[t, x],
                        blk.gas_solid_htc_eqn[t, x],
                        **calc_var_kwds,
                    )
                    calculate_variable_from_constraint(
                        blk.gas_phase.heat[t, x],
                        blk.gas_phase_heat_transfer[t, x],
                        **calc_var_kwds,
                    )
                    calculate_variable_from_constraint(
                        blk.solid_phase_heat[t, x],
                        blk.solid_phase_heat_transfer[t, x],
                        **calc_var_kwds,
                    )

                    if t != blk.flowsheet().time.first():
                        calculate_variable_from_constraint(
                            blk.gas_phase.energy_accumulation[t, x, "Vap"],
                            blk.gas_phase.energy_accumulation_disc_eq[t, x, "Vap"],
                            **calc_var_kwds,
                        )
                        calculate_variable_from_constraint(
                            blk.solid_energy_accumulation[t, x],
                            blk.solid_energy_accumulation_disc_eq[t, x],
                            **calc_var_kwds,
                        )

            blk.reynolds_number_particle.activate()
            blk.prandtl_number.activate()
            blk.nusselt_number_particle.activate()
            blk.gas_solid_htc_eqn.activate()

            # Activate gas phase energy balance equations
            blk.gas_phase_heat_transfer.activate()
            blk.gas_phase.enthalpy_balances.activate()

            # Activate solid phase energy balance equations
            blk.solid_phase_heat_transfer.activate()
            blk.solid_enthalpy_balances.activate()

            init_log.info("Initialize Energy Balances")
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee, symbolic_solver_labels=True)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 4 {}.".format(idaeslog.condition(results))
                )
            else:
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

        # Initialize energy balance
        if blk.config.energy_balance_type == EnergyBalanceType.none:
            if blk.config.flow_type == "forward_flow":
                x_inlet = blk.length_domain.first()
            else:
                x_inlet = blk.length_domain.last()
            t0 = blk.flowsheet().config.time.first()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    if t != blk.flowsheet().config.time.first() and x != x_inlet:
                        # Unfix gas temperature variables except at the inlet
                        blk.gas_phase.properties[t, x].temperature.unfix()
                        blk.gas_phase.properties[t, x].temperature = value(
                            blk.solid_properties[t0, x].temperature
                        )

            for t in blk.flowsheet().time:
                t0 = blk.flowsheet().config.time.first()
                if t != blk.flowsheet().config.time.first():
                    for x in blk.length_domain:
                        # Unfix solid temperature variables except initial
                        blk.solid_properties[t, x].temperature = value(
                            blk.solid_properties[t0, x].temperature
                        )
                        blk.solid_properties[t, x].temperature.unfix()

            blk.isothermal_gas_phase.activate()
            blk.isothermal_solid_phase.activate()

            init_log.info("Initialize Energy Balances")
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 4 {}.".format(idaeslog.condition(results))
                )
            else:
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

    def block_triangularization_initialize(
        blk,
        gas_phase_state_args=None,
        solid_phase_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        calc_var_kwds=None,
    ):
        """
        Block triangularization (BT) initialization routine for 1D FixedBed unit.

        The BT initialization routine initializes the unit model in the following steps:

        Step 1: Convert the system of equations from a partial differential set of
                equations (PDAE) to an algebraic set of equations (AE) by deactivating
                the discretization equations and sum_component_equations (if present),
                and fixing the state variables to a guess (typically inlet conditions).
        Step 2: Decompose the AE into strongly connected components
                via block triangularization and solve individual blocks
        Step 3: revert the fixed variables and deactivated constraints to their
                original states.

        The initialization procedure is wrapped in the TemporarySubsystemManager utility
        to ensure that the deactivation of constraints and fixing of variables are only
        temporary and the final model structure is unchanged after initialization.

        Keyword Arguments:
            gas_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            solid_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            calc_var_kwds: Dictionary of keyword arguments for
                calculate_variable_from_constraint

        Returns:
            None
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")

        # =========================================================================
        # Check that the original model has zero degrees of freedom
        if degrees_of_freedom(blk) != 0:
            msg = (
                "Original model has nonzero degrees of freedom. This was "
                "unexpected. "
                + "\n"
                + "Degrees of freedom: "
                + str(degrees_of_freedom(blk))
            )
            init_log.error(msg)
            raise ValueError("Nonzero degrees of freedom.")

        # =========================================================================
        # Create list of constraints to deactivate
        to_deactivate = []
        # Deactivate gas/solid phase sum equations and discretization equations
        # (mass/energy flow, mass/energy accumulation, pressure)
        endswith_terms = ("disc_eq", "sum_component_eqn")
        for c in blk.component_objects(Constraint, descend_into=True):
            if c.local_name.endswith(endswith_terms):
                c.attribute_errors_generate_exceptions = False
                to_deactivate.append(c)

        # Additional equations to deactivate if energy balance type is none
        if hasattr(blk, "isothermal_gas_phase"):
            isothermal_gas_phase_eqn = blk.isothermal_gas_phase[...]
            isothermal_gas_phase_eqn.attribute_errors_generate_exceptions = False
            to_deactivate.extend(isothermal_gas_phase_eqn)
        if hasattr(blk, "isothermal_solid_phase"):
            isothermal_solid_phase_eqn = blk.isothermal_solid_phase[...]
            isothermal_solid_phase_eqn.attribute_errors_generate_exceptions = False
            to_deactivate.extend(isothermal_solid_phase_eqn)

        # =========================================================================
        # Variables to fix during initialization

        # List of state vars to fix (typically fixed to boundary/initial conditions)
        gas_flags = fix_state_vars(blk.gas_phase.properties, gas_phase_state_args)
        solid_flags = fix_state_vars(blk.solid_properties, solid_phase_state_args)

        # List of additional variables to fix
        to_fix = []
        # Flow derivative variables are fixed to ensure square AE
        set_value = 0
        time_set = blk.flowsheet().time
        for t in blk.flowsheet().time:
            for var in ComponentSet(blk.component_objects(Var)):
                if var.name.endswith("_flow_dx"):
                    n = var.index_set().dimen
                    if n == 1:
                        var.set_value(set_value)
                        to_fix.append(var[t])
                    elif n >= 2:
                        info = get_index_set_except(var, time_set)
                        non_time_set = info["set_except"]
                        index_getter = info["index_getter"]
                        for non_time_index in non_time_set:
                            index = index_getter(non_time_index, t)
                            var[index].set_value(set_value)
                            to_fix.append(var[index])

        # =========================================================================
        with TemporarySubsystemManager(to_fix=to_fix, to_deactivate=to_deactivate):

            # Check if the system is structurally singular
            igraph = IncidenceGraphInterface(blk)
            N = len(igraph.constraints)
            M = len(igraph.variables)
            matching = igraph.maximum_matching()
            if not (N == len(matching) and M == len(matching)):
                msg = (
                    "Model is structurally singular as maximum matching "
                    "does not include all costraints and variables "
                    "\n"
                    "Number of constraints: "
                    + str(N)
                    + "\n"
                    + "Number of variables: "
                    + str(M)
                    + "\n"
                    + "Maximum matching: "
                    + str(len(matching))
                )
                init_log.warning(msg)
                raise ValueError("Structural singularity")

            # Decompose AE system into strongly connected components and solve
            if calc_var_kwds is None:
                calc_var_kwds = {"eps": 1e-8}
            try:
                solve_strongly_connected_components(
                    blk, calc_var_kwds=calc_var_kwds, solver=solver
                )
            except RuntimeError:
                msg = (
                    "Initial value for variable results in a derivative value "
                    "that is very close to zero. Please provide a different initial "
                    "guess and/or adjust calc_var_kwds tolerance setting."
                )
                init_log.warning(msg)
                raise

        # Revert the state vars to their original state
        revert_state_vars(blk.gas_phase.properties, gas_flags)
        revert_state_vars(blk.solid_properties, solid_flags)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # local aliases used to shorten object names
        gas_phase = self.config.gas_phase_config
        solid_phase = self.config.solid_phase_config

        # Get units meta data from property packages
        units_meta_gas = gas_phase.property_package.get_metadata().get_derived_units

        # scale some variables
        if hasattr(self, "bed_diameter"):
            sf = 1 / value(self.bed_diameter)
            iscale.set_scaling_factor(self.bed_diameter, 1e-1 * sf)

        if hasattr(self, "bed_area"):
            sf = 1 / value(constants.pi * (0.5 * self.bed_diameter) ** 2)
            iscale.set_scaling_factor(self.bed_area, 1e-1 * sf)

        if hasattr(self, "solid_phase_area"):
            sf = iscale.get_scaling_factor(self.bed_area)
            iscale.set_scaling_factor(self.solid_phase_area, sf)

        if hasattr(self.gas_phase, "area"):
            sf = iscale.get_scaling_factor(self.bed_area)
            iscale.set_scaling_factor(self.gas_phase.area, sf)

        if self.config.energy_balance_type != EnergyBalanceType.none:
            if hasattr(self, "Re_particle"):
                for (t, x), v in self.Re_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf1 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].dens_mass
                        )
                        sf2 = 1 / value(
                            pyunits.convert(
                                self.solid_properties[t, x].params.particle_dia,
                                to_units=units_meta_gas("length"),
                            )
                        )
                        iscale.set_scaling_factor(v, 1e-1 * sf1 * sf2)

        if hasattr(self, "solid_material_accumulation"):
            for (t, x, j), v in self.solid_material_accumulation.items():
                if iscale.get_scaling_factor(v) is None:
                    sf = iscale.get_scaling_factor(
                        self.solid_properties[t, x].get_material_density_terms(
                            "Sol", j
                        ),
                        default=1,
                        warning=True,
                    )
                    iscale.set_scaling_factor(v, sf)

        if self.config.energy_balance_type != EnergyBalanceType.none:
            if hasattr(self, "Pr_particle"):
                for (t, x), v in self.Pr_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf1 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].cp_mass
                        )
                        sf2 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].visc_d
                        )
                        iscale.set_scaling_factor(v, sf1 * sf2)

            if hasattr(self, "Nu_particle"):
                for (t, x), v in self.Nu_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf = 1 / (
                            value(
                                (
                                    (
                                        2.0
                                        + 1.1
                                        * (
                                            smooth_abs(self.Re_particle[t, x], self.eps)
                                            ** 0.6
                                        )
                                        ** 3
                                    )
                                    * self.Pr_particle[t, x]
                                )
                                ** (1 / 3)
                            )
                        )
                        iscale.set_scaling_factor(v, 1e-2 * sf)

            if hasattr(self, "gas_solid_htc"):
                for (t, x), v in self.gas_solid_htc.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf1 = iscale.get_scaling_factor(self.Nu_particle[t, x])
                        sf2 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].therm_cond
                        )
                        iscale.set_scaling_factor(v, sf1 * sf2)

            if hasattr(self, "solid_energy_accumulation"):
                for (t, x), v in self.solid_energy_accumulation.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf = iscale.get_scaling_factor(
                            self.solid_properties[t, x].get_energy_density_terms("Sol"),
                            default=1,
                            warning=True,
                        )
                        iscale.set_scaling_factor(v, sf)

            if hasattr(self, "solid_energy_holdup"):
                for (t, x), v in self.solid_energy_holdup.items():
                    sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                    sf2 = iscale.get_scaling_factor(
                        self.solid_properties[t, x].get_energy_density_terms("Sol")
                    )
                    iscale.set_scaling_factor(v, sf1 * sf2)

            # scale heat variables in CV1D
            for (t, x), v in self.gas_phase.heat.items():
                sf1 = iscale.get_scaling_factor(
                    self.gas_phase.properties[t, x].enth_mol
                )
                sf2 = iscale.get_scaling_factor(
                    self.gas_phase.properties[t, x].flow_mol
                )
                iscale.set_scaling_factor(v, sf1 * sf2)

            for (t, x), v in self.solid_phase_heat.items():
                sf1 = iscale.get_scaling_factor(self.solid_properties[t, x].enth_mass)
                sf2 = iscale.get_scaling_factor(
                    self.solid_properties[t, x].dens_mass_particle
                )
                sf3 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                iscale.set_scaling_factor(v, sf1 * sf2 * sf3)

        # Scale some constraints
        if hasattr(self, "bed_area_eqn"):
            for c in self.bed_area_eqn.values():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "gas_phase_area_constraint"):
            for (t, x), c in self.gas_phase_area_constraint.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "solid_phase_area_constraint"):
            for (t, x), c in self.solid_phase_area_constraint.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "gas_super_vel"):
            for (t, x), c in self.gas_super_vel.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.gas_phase.properties[t, x].flow_mol),
                    overwrite=False,
                )

        if hasattr(self, "gas_phase_config_pressure_drop"):
            for (t, x), c in self.gas_phase_config_pressure_drop.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.gas_phase.deltaP[t, x]),
                    overwrite=False,
                )

        if hasattr(self, "gas_phase_config_rxn_ext"):
            for (t, x, r), c in self.gas_phase_config_rxn_ext.items():
                sf1 = iscale.get_scaling_factor(
                    self.gas_phase.reactions[t, x].reaction_rate[r]
                )
                sf2 = iscale.get_scaling_factor(self.bed_area)
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "gas_comp_hetero_rxn"):
            for (t, x, p, j), c in self.gas_comp_hetero_rxn.items():
                iscale.constraint_scaling_transform(
                    c,
                    1e-3
                    * iscale.get_scaling_factor(
                        self.gas_phase.mass_transfer_term[t, x, p, j]
                    ),
                    overwrite=False,
                )

        if self.config.energy_balance_type != EnergyBalanceType.none:
            if hasattr(self, "reynolds_number_particle"):
                for (t, x), c in self.reynolds_number_particle.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Re_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "prandtl_number"):
                for (t, x), c in self.prandtl_number.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Pr_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "nusselt_number_particle"):
                for (t, x), c in self.nusselt_number_particle.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Nu_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "gas_solid_htc_eqn"):
                for (t, x), c in self.gas_solid_htc_eqn.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.gas_solid_htc[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "gas_phase_heat_transfer"):
                for (t, x), c in self.gas_phase_heat_transfer.items():
                    iscale.constraint_scaling_transform(
                        c,
                        1e-2 * iscale.get_scaling_factor(self.gas_phase.heat[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "solid_phase_heat_transfer"):
                for (t, x), c in self.solid_phase_heat_transfer.items():
                    iscale.constraint_scaling_transform(
                        c,
                        1e-2 * iscale.get_scaling_factor(self.solid_phase_heat[t, x]),
                        overwrite=False,
                    )

        elif self.config.energy_balance_type == EnergyBalanceType.none:
            if hasattr(self, "isothermal_gas_phase"):
                for (t, x), c in self.isothermal_gas_phase.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].temperature
                        ),
                        overwrite=False,
                    )

            if hasattr(self, "isothermal_solid_phase"):
                for (t, x), c in self.isothermal_solid_phase.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(
                            self.solid_properties[t, x].temperature
                        ),
                        overwrite=False,
                    )

        # Scale some constraints
        if hasattr(self, "solid_material_holdup_calculation"):
            for (t, x, j), c in self.solid_material_holdup_calculation.items():
                sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                sf2 = iscale.get_scaling_factor(
                    self.solid_properties[t, x].dens_mass_particle
                )
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "solid_material_accumulation_disc_eq"):
            for (t, x, j), c in self.solid_material_accumulation_disc_eq.items():
                iscale.constraint_scaling_transform(
                    c,
                    1e-6
                    * iscale.get_scaling_factor(
                        self.solid_material_accumulation[t, x, j]
                    ),
                    overwrite=False,
                )

        if hasattr(self, "solid_material_balances"):
            component_list = solid_phase.property_package.component_list
            # Get a single representative value in component list
            for j in component_list:
                break
            if solid_phase.reaction_package is not None:
                rate_reaction_idx = solid_phase.reaction_package.rate_reaction_idx
                # Get a single representative value in rate_reaction index list
                for r in rate_reaction_idx:
                    break
                for (t, x, j), c in self.solid_material_balances.items():
                    sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                    sf2 = iscale.get_scaling_factor(
                        self.solid_properties[t, x].params.mw_comp[j],
                        default=1e1,
                    )
                    sf3 = iscale.get_scaling_factor(
                        self.solid_reactions[t, x].reaction_rate[r]
                    )
                    iscale.constraint_scaling_transform(
                        c, sf1 * sf2 * sf3, overwrite=False
                    )
            else:
                for (t, x, j), c in self.solid_material_balances.items():
                    sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                    sf2 = iscale.get_scaling_factor(
                        self.solid_properties[t, x].params.mw_comp[j],
                        default=1e1,
                    )
                    iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "solid_sum_component_eqn"):
            component_list = solid_phase.property_package.component_list
            # Get a single representative value in component list
            for j in component_list:
                break
            for (t, x), c in self.solid_sum_component_eqn.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.solid_properties[t, x].mass_frac_comp[j]
                    ),
                    overwrite=False,
                )

        if self.config.energy_balance_type != EnergyBalanceType.none:

            if hasattr(self, "solid_energy_holdup_calculation"):
                for (t, x), c in self.solid_energy_holdup_calculation.items():
                    sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                    sf2 = iscale.get_scaling_factor(
                        self.solid_properties[t, x].get_energy_density_terms("Sol")
                    )
                    iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

            if hasattr(self, "solid_enthalpy_balances"):
                for (t, x), c in self.solid_enthalpy_balances.items():
                    sf1 = iscale.get_scaling_factor(self.solid_phase_area[t, x])
                    sf2 = iscale.get_scaling_factor(
                        self.solid_properties[t, x].enth_mass
                    )
                    iscale.constraint_scaling_transform(
                        c, 1e-2 * sf1 * sf2, overwrite=False
                    )
            if hasattr(self.gas_phase, "enthalpy_balances"):
                phase_list = self.gas_phase.properties.phase_list
                for (t, x), c in self.gas_phase.enthalpy_balances.items():
                    # This is OK
                    sf = iscale.min_scaling_factor(
                        [
                            # pylint: disable-next=protected-access
                            self.gas_phase._enthalpy_flow[t, x, p]
                            for p in phase_list
                        ]
                    )
                    iscale.constraint_scaling_transform(c, 1e-3 * sf, overwrite=True)

            if hasattr(self, "solid_energy_accumulation_disc_eq"):
                for (t, x), c in self.solid_energy_accumulation_disc_eq.items():
                    iscale.constraint_scaling_transform(
                        c,
                        1e-3
                        * iscale.get_scaling_factor(
                            self.solid_energy_accumulation[t, x]
                        ),
                        overwrite=False,
                    )

            if hasattr(self.gas_phase, "energy_accumulation_disc_eq"):
                for (t, x, p), c in self.gas_phase.energy_accumulation_disc_eq.items():
                    iscale.constraint_scaling_transform(
                        c,
                        1e-3
                        * iscale.get_scaling_factor(
                            self.gas_phase.energy_accumulation[t, x, p]
                        ),
                        overwrite=True,
                    )

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {"Gas Inlet": self.gas_inlet, "Gas Outlet": self.gas_outlet},
            time_point=time_point,
        )

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Bed Height"] = self.bed_height
        var_dict["Bed Area"] = self.bed_area
        var_dict["Gas Inlet Velocity"] = self.velocity_superficial_gas[time_point, 0]
        var_dict["Gas Outlet Velocity"] = self.velocity_superficial_gas[time_point, 1]

        return {"vars": var_dict}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Steam turbine stage model. This is a standard isentropic turbine. Under off-design
conditions the base efficiency and pressure ratio do not change much for the stages
between the inlet and outlet. This model is based on:

Liese, (2014). "Modeling of a Steam Turbine Including Partial Arc Admission
    for Use in a Process Simulation Software Environment." Journal of Engineering
    for Gas Turbines and Power. v136.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

__Author__ = "John Eslick"

from pyomo.environ import Var, units as pyunits

from idaes.core import declare_process_block_class
from idaes.models_extra.power_generation.unit_models.helm.turbine import (
    HelmIsentropicTurbineData,
)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("HelmTurbineStage", doc="Basic steam turbine model")
class HelmTurbineStageData(HelmIsentropicTurbineData):
    CONFIG = HelmIsentropicTurbineData.CONFIG()

    def build(self):
        super().build()

        self.efficiency_mech = Var(initialize=1.0, doc="Turbine mechanical efficiency")
        self.efficiency_mech.fix()
        time_set = self.flowsheet().time
        self.shaft_speed = Var(
            time_set, doc="Shaft speed [1/s]", initialize=60.0, units=pyunits.s**-1
        )
        self.shaft_speed.fix()

        @self.Expression(time_set, doc="Specific speed [dimensionless]")
        def specific_speed(b, t):
            s = b.shaft_speed[t]  # 1/s
            v = b.control_volume.properties_out[t].flow_vol  # m3/s
            his_rate = b.work_isentropic[t]  # J/s
            m = b.control_volume.properties_out[t].flow_mass  # kg/s
            return s * v**0.5 * (his_rate / m) ** (-0.75)  # dimensionless

        @self.Expression(time_set, doc="Thermodynamic power [J/s]")
        def power_thermo(b, t):
            return b.control_volume.work[t]

        @self.Expression(self.flowsheet().time, doc="Shaft power [J/s]")
        def power_shaft(b, t):
            return b.power_thermo[t] * b.efficiency_mech

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialize the turbine stage model.  This deactivates the
        specialized constraints, then does the isentropic turbine initialization,
        then reactivates the constraints and solves.

        Args:
            outlvl : sets output level of initialization routine
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """
        super().initialize_build(outlvl=outlvl, solver=solver, optarg=optarg)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Natural gas property package for the vapor phase using Peng-Robinson equation
of state.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# Import Python libraries
import logging
import copy
import enum

# Import Pyomo units
import pyomo.environ as pyo
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import VaporPhase, LiquidPhase, Component, PhaseType

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.phase_equil.forms import (
    log_fugacity,
)
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.pure import (
    NIST,
    RPP4,
    RPP5,
    ChapmanEnskogLennardJones,
    Eucken,
)
from idaes.models.properties.modular_properties.base.generic_reaction import (
    ConcentrationForm,
)
from idaes.models.properties.modular_properties.transport_properties import (
    ViscosityWilke,
    ThermalConductivityWMS,
    NoMethod,
)
from idaes.models.properties.modular_properties.transport_properties.viscosity_wilke import (
    wilke_phi_ij_callback,
)

from idaes.models.properties.modular_properties.reactions.dh_rxn import constant_dh_rxn
from idaes.models.properties.modular_properties.reactions.rate_constant import arrhenius
from idaes.models.properties.modular_properties.reactions.rate_forms import (
    power_law_rate,
)
from idaes.core.util.exceptions import ConfigurationError

# Set up logger
_log = logging.getLogger(__name__)


class EosType(enum.Enum):
    PR = 1
    IDEAL = 2


# Property Sources

# Source: NIST webbook
# Properties: Heat capacity coefficients for all species except ethane,
# propane, and butane. Reference enthalpies and entropies for all species.

# Source: The Properties of Gases and Liquids (1987)
# 4th edition, Chemical Engineering Series - Robert C. Reid
# Properties: Critical temperatures and pressures. Omega.
# Heat capacity coefficients for ethane, propane, and butane.

_phase_dicts_pr = {
    "Vap": {
        "type": VaporPhase,
        "equation_of_state": Cubic,
        "equation_of_state_options": {"type": CubicType.PR},
        "visc_d_phase": ViscosityWilke,
        "therm_cond_phase": ThermalConductivityWMS,
    },
    "Liq": {
        "type": LiquidPhase,
        "equation_of_state": Cubic,
        "equation_of_state_options": {"type": CubicType.PR},
        "visc_d_phase": NoMethod,
        "therm_cond_phase": NoMethod,
    },
}

_phase_dicts_ideal = {
    "Vap": {
        "type": VaporPhase,
        "equation_of_state": Ideal,
        "visc_d_phase": ViscosityWilke,
        "transport_property_options": {
            "viscosity_phi_ij_callback": wilke_phi_ij_callback,
        },
        "therm_cond_phase": ThermalConductivityWMS,
    },
}

_component_params = {
    "H2": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"H": 2},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0020159, pyunits.kg / pyunits.mol),
            "pressure_crit": (13e5, pyunits.Pa),
            "temperature_crit": (33.2, pyunits.K),
            "omega": -0.218,
            "cp_mol_ig_comp_coeff": {
                "A": 33.066178,
                "B": -11.363417,
                "C": 11.432816,
                "D": -2.772874,
                "E": -0.158558,
                "F": -9.980797,
                "G": 172.707974,
                "H": 0.0,
            },
            "lennard_jones_sigma": (2.826, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (59.7, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "CO": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 1, "O": 1},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0280101, pyunits.kg / pyunits.mol),
            "pressure_crit": (35e5, pyunits.Pa),
            "temperature_crit": (132.9, pyunits.K),
            "omega": 0.066,
            "cp_mol_ig_comp_coeff": {
                "A": 25.56759,
                "B": 6.09613,
                "C": 4.054656,
                "D": -2.671301,
                "E": 0.131021,
                "F": -118.0089,
                "G": 227.3665,
                "H": -110.5271,
            },
            "lennard_jones_sigma": (3.690, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (91.7, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "H2O": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase, PhaseType.liquidPhase],
        "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
        "elemental_composition": {"H": 2, "O": 1},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "pressure_sat_comp": NIST,
        "parameter_data": {
            "mw": (0.01801528, pyunits.kg / pyunits.mol),
            "pressure_crit": (221.2e5, pyunits.Pa),
            "temperature_crit": (647.3, pyunits.K),
            "omega": 0.344,
            "cp_mol_ig_comp_coeff": {
                "A": 30.092,
                "B": 6.832514,
                "C": 6.793435,
                "D": -2.53448,
                "E": 0.082139,
                "F": -250.881,
                "G": 223.3967,
                "H": -241.8264,
            },
            "pressure_sat_comp_coeff": {  # NIST <- Stull 1947
                "A": 4.6543,
                "B": 1435.264,
                "C": -64.848,
            },
            "lennard_jones_sigma": (2.641, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (809.1, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "CO2": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "elemental_composition": {"C": 1, "O": 2},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "parameter_data": {
            "mw": (0.04401, pyunits.kg / pyunits.mol),
            "pressure_crit": (73.8e5, pyunits.Pa),
            "temperature_crit": (304.1, pyunits.K),
            "omega": 0.239,
            "cp_mol_ig_comp_coeff": {
                "A": 24.99735,
                "B": 55.18696,
                "C": -33.69137,
                "D": 7.948387,
                "E": -0.136638,
                "F": -403.6075,
                "G": 228.2431,
                "H": -393.5224,
            },
            "lennard_jones_sigma": (3.941, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (195.2, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "O2": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "elemental_composition": {"O": 2},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "parameter_data": {
            "mw": (0.031998, pyunits.kg / pyunits.mol),
            "pressure_crit": (50.4e5, pyunits.Pa),
            "temperature_crit": (154.6, pyunits.K),
            "omega": 0.025,
            "cp_mol_ig_comp_coeff": {
                "A": 30.03235,
                "B": 8.772972,
                "C": -3.988133,
                "D": 0.788313,
                "E": -0.741599,
                "F": -11.32468,
                "G": 236.1663,
                "H": 0.0,
            },
            "lennard_jones_sigma": (3.467, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (106.7, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "N2": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"N": 2},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0280134, pyunits.kg / pyunits.mol),
            "pressure_crit": (33.9e5, pyunits.Pa),
            "temperature_crit": (126.2, pyunits.K),
            "omega": 0.039,
            "cp_mol_ig_comp_coeff": {
                "A": 19.50583,
                "B": 19.88705,
                "C": -8.598535,
                "D": 1.369784,
                "E": 0.527601,
                "F": -4.935202,
                "G": 212.39,
                "H": 0.0,
            },
            "lennard_jones_sigma": (3.798, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (71.4, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "Ar": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"Ar": 1},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.039948, pyunits.kg / pyunits.mol),
            "pressure_crit": (48.7e5, pyunits.Pa),
            "temperature_crit": (150.8, pyunits.K),
            "omega": 0.001,
            "cp_mol_ig_comp_coeff": {
                "A": 20.786,
                "B": 2.825911e-07,
                "C": -1.464191e-07,
                "D": 1.092131e-08,
                "E": -3.661371e-08,
                "F": -6.19735,
                "G": 179.999,
                "H": 0.0,
            },
            "lennard_jones_sigma": (3.542, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (93.3, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "CH4": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 1, "H": 4},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0160425, pyunits.kg / pyunits.mol),
            "pressure_crit": (46e5, pyunits.Pa),
            "temperature_crit": (190.4, pyunits.K),
            "omega": 0.011,
            "cp_mol_ig_comp_coeff": {
                "A": -0.703029,
                "B": 108.4773,
                "C": -42.52157,
                "D": 5.862788,
                "E": 0.678565,
                "F": -76.84376,
                "G": 158.7163,
                "H": -74.8731,
            },
            "lennard_jones_sigma": (3.758, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (148.6, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "C2H6": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 2, "H": 6},
        "enth_mol_ig_comp": RPP4,
        "entr_mol_ig_comp": RPP4,
        "cp_mol_ig_comp": RPP4,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.030069, pyunits.kg / pyunits.mol),
            "pressure_crit": (48.8e5, pyunits.Pa),
            "temperature_crit": (305.4, pyunits.K),
            "omega": 0.099,
            "cp_mol_ig_comp_coeff": {
                "A": 5.409,
                "B": 1.781e-1,
                "C": -6.938e-5,
                "D": 8.713e-9,
            },
            "enth_mol_form_vap_comp_ref": (-84000, pyunits.J / pyunits.mol),
            "entr_mol_form_vap_comp_ref": (229.2, pyunits.J / pyunits.mol / pyunits.K),
            "lennard_jones_sigma": (4.443, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (215.7, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "C3H8": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 3, "H": 8},
        "enth_mol_ig_comp": RPP4,
        "entr_mol_ig_comp": RPP4,
        "cp_mol_ig_comp": RPP4,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0320849, pyunits.kg / pyunits.mol),
            "pressure_crit": (42.5e5, pyunits.Pa),
            "temperature_crit": (369.8, pyunits.K),
            "omega": 0.153,
            "cp_mol_ig_comp_coeff": {
                "A": -4.224,
                "B": 3.063e-1,
                "C": -1.586e-4,
                "D": 3.215e-8,
            },
            "enth_mol_form_vap_comp_ref": (-104700, pyunits.J / pyunits.mol),
            "entr_mol_form_vap_comp_ref": (270.3, pyunits.J / pyunits.mol / pyunits.K),
            "lennard_jones_sigma": (5.118, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (237.1, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "C4H10": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 4, "H": 10},
        "enth_mol_ig_comp": RPP5,
        "entr_mol_ig_comp": RPP5,
        "cp_mol_ig_comp": RPP5,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.058123, pyunits.kg / pyunits.mol),  # RPP5
            "pressure_crit": (37.96e5, pyunits.Pa),  # RPP5
            "temperature_crit": (425.12, pyunits.K),  # RPP5
            "omega": 0.200,  # RPP5
            "cp_mol_ig_comp_coeff": {  # RPP5
                "a0": 5.547,
                "a1": 5.536e-3,
                "a2": 8.057e-5,
                "a3": -10.571e-8,
                "a4": 4.134e-11,
            },
            "enth_mol_form_vap_comp_ref": (-125790, pyunits.J / pyunits.mol),  # RPP5
            "entr_mol_form_vap_comp_ref": (
                310.23,
                pyunits.J / pyunits.mol / pyunits.K,
            ),  # wikipedia data page
            # LJ parameters for n-butane from Appendix B of Properties of Gases and Liquids, 5th Ed.
            # Estimating LJ parameters from viscosity is underdetermined: the group
            # sigma ** 2 * sqrt(epsilon) is more significant. For n-butane, that is 506.4 angstrom ** 2 sqrt(K)
            # for the LJ parameters for isobutane, it is 506.1 angstrom ** 2 sqrt(K).
            # In short, the n-butane parameters will adequately represent both species for most temperatures
            "lennard_jones_sigma": (4.687, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (531.4, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "H2S": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"H": 2, "S": 1},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.034081, pyunits.kg / pyunits.mol),  # NIST
            "pressure_crit": (89.6291e5, pyunits.Pa),  # NIST <- Goodwin 1983
            "temperature_crit": (373.4, pyunits.K),  # NIST <- Goodwin 1983
            "omega": 0.090,  # RPP5
            "cp_mol_ig_comp_coeff": {  # NIST <- Chase 1998
                "A": 26.88412,
                "B": 18.67809,
                "C": 3.434203,
                "D": -3.378702,
                "E": 0.135882,
                "F": -28.91211,
                "G": 233.3747,
                "H": -20.50202,
            },
            "lennard_jones_sigma": (3.623, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (301.1, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "SO2": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"S": 1, "O": 2},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.064064, pyunits.kg / pyunits.mol),  # NIST
            "pressure_crit": (78.84e5, pyunits.Pa),  # RPP5
            "temperature_crit": (430.80, pyunits.K),  # RPP5
            "omega": 0.0,  # RPP5 (Not listed)
            "cp_mol_ig_comp_coeff": {  # NIST <- Chase 1998
                "A": 21.43049,
                "B": 74.35094,
                "C": -57.75217,
                "D": 16.35534,
                "E": 0.086731,
                "F": -305.7688,
                "G": 254.8872,
                "H": -296.8422,
            },
            "lennard_jones_sigma": (4.112, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (335.4, pyunits.K),
            "f_int_eucken": 1,
        },
    },
    "C2H4": {
        "type": Component,
        "valid_phase_types": [PhaseType.vaporPhase],
        "elemental_composition": {"C": 2, "H": 4},
        "enth_mol_ig_comp": NIST,
        "entr_mol_ig_comp": NIST,
        "cp_mol_ig_comp": NIST,
        "visc_d_phase_comp": {"Vap": ChapmanEnskogLennardJones},
        "therm_cond_phase_comp": {"Vap": Eucken},
        "parameter_data": {
            "mw": (0.0280532, pyunits.kg / pyunits.mol),  # NIST
            "pressure_crit": (50.6e5, pyunits.Pa),  # NIST
            "temperature_crit": (282.5, pyunits.K),  # NIST
            "omega": 0.087,  # RPP5
            "cp_mol_ig_comp_coeff": {  # NIST <- Chase 1998
                "A": -6.387880,
                "B": 184.4019,
                "C": -112.9718,
                "D": 28.49593,
                "E": 0.315540,
                "F": 48.17332,
                "G": 163.1568,
                "H": 52.46694,
            },
            "lennard_jones_sigma": (4.163, pyunits.angstrom),
            "lennard_jones_epsilon_reduced": (224.7, pyunits.K),
            "f_int_eucken": 1,
        },
    },
}

_water_visc_d = {"Vap": ChapmanEnskogLennardJones, "Liq": None}
_water_therm_cond = {"Vap": Eucken, "Liq": None}


# returns a configuration dictionary for the list of specified components
def get_prop(components=None, phases="Vap", eos=EosType.PR, scaled=False):
    if components is None:
        components = list(_component_params.keys())
    configuration = {
        "components": {},  # fill in later based on selected components
        "parameter_data": {},
        "phases": {},
        # Set base units of measurement
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        # Specifying state definition
        "state_definition": FTPx,
        "state_bounds": {
            "flow_mol": (0, 8000, 50000, pyunits.mol / pyunits.s),
            "temperature": (273.15, 500, 2500, pyunits.K),
            "pressure": (5e4, 1.3e5, 1e8, pyunits.Pa),
        },
        "pressure_ref": (101325, pyunits.Pa),
        "temperature_ref": (298.15, pyunits.K),
    }

    c = configuration["components"]
    if isinstance(phases, str):
        phases = [phases]
    for comp in components:
        c[comp] = copy.deepcopy(_component_params[comp])
        if comp == "H2O":
            c["H2O"]["visc_d_phase_comp"] = copy.deepcopy(
                {p: _water_visc_d[p] for p in phases}
            )
            c["H2O"]["therm_cond_phase_comp"] = copy.deepcopy(
                {p: _water_therm_cond[p] for p in phases}
            )
    for k in phases:
        if eos == EosType.PR:
            configuration["phases"][k] = copy.deepcopy(_phase_dicts_pr[k])
        elif eos == EosType.IDEAL:
            if k == "Liq":
                raise ConfigurationError(
                    "This parameter set does not support Ideal EOS with liquid"
                )
            configuration["phases"][k] = copy.deepcopy(_phase_dicts_ideal[k])
        else:
            raise ValueError("Invalid EoS.")
    if len(phases) > 1:
        p = tuple(phases)
        configuration["phases_in_equilibrium"] = [p]
        configuration["phase_equilibrium_state"] = {p: SmoothVLE}

    # Fill the binary parameters with zeros.
    d = configuration["parameter_data"]
    d["PR_kappa"] = {(a, b): 0 for a in c for b in c}

    # Change to scaled units if specified
    if scaled:
        configuration["base_units"]["mass"] = pyunits.Mg
        configuration["base_units"]["amount"] = pyunits.kmol

    return configuration


def get_rxn(property_package, reactions=None, scaled=False):
    rxns = {
        "property_package": property_package,
        "base_units": {
            "time": pyo.units.s,
            "length": pyo.units.m,
            "mass": pyo.units.kg,
            "amount": pyo.units.mol,
            "temperature": pyo.units.K,
        },
        "rate_reactions": {
            "ch4_cmb": {
                "stoichiometry": {
                    ("Vap", "CH4"): -1,
                    ("Vap", "O2"): -2,
                    ("Vap", "H2O"): 2,
                    ("Vap", "CO2"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "c2h6_cmb": {
                "stoichiometry": {
                    ("Vap", "C2H6"): -2,
                    ("Vap", "O2"): -7,
                    ("Vap", "H2O"): 6,
                    ("Vap", "CO2"): 4,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "c3h8_cmb": {
                "stoichiometry": {
                    ("Vap", "C3H8"): -1,
                    ("Vap", "O2"): -5,
                    ("Vap", "H2O"): 4,
                    ("Vap", "CO2"): 3,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "c4h10_cmb": {
                "stoichiometry": {
                    ("Vap", "C4H10"): -2,
                    ("Vap", "O2"): -13,
                    ("Vap", "H2O"): 10,
                    ("Vap", "CO2"): 8,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "co_cmb": {
                "stoichiometry": {
                    ("Vap", "CO"): -2,
                    ("Vap", "O2"): -1,
                    ("Vap", "CO2"): 2,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "h2_cmb": {
                "stoichiometry": {
                    ("Vap", "H2"): -2,
                    ("Vap", "O2"): -1,
                    ("Vap", "H2O"): 2,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "h2s_cmb": {
                "stoichiometry": {
                    ("Vap", "H2S"): -2,
                    ("Vap", "O2"): -3,
                    ("Vap", "H2O"): 2,
                    ("Vap", "SO2"): 2,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
            "c2h4_cmb": {
                "stoichiometry": {
                    ("Vap", "C2H4"): -1,
                    ("Vap", "O2"): -3,
                    ("Vap", "H2O"): 2,
                    ("Vap", "CO2"): 2,
                },
                "heat_of_reaction": constant_dh_rxn,
                "rate_constant": arrhenius,
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": 0,
                    "arrhenius_const": 0,
                    "energy_activation": 0,
                },
            },
        },
    }
    # Change to scaled units if specified
    if scaled:
        rxns["base_units"]["mass"] = pyunits.Mg
        rxns["base_units"]["amount"] = pyunits.kmol
    if reactions is None:
        return rxns
    else:
        for r in list(rxns["rate_reactions"].keys()):
            if r not in reactions:
                del rxns["rate_reactions"][r]
        return rxns



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for methods from electrolyte library

Authors: Andrew Lee
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.electrolyte import (
    relative_permittivity_constant,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {"relative_permittivity_liq_comp": 101}

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=298.15, units=pyunits.K)
    m.params.pressure_ref = Var(initialize=1e5, units=pyunits.Pa)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=500, units=pyunits.K)
    m.props[1].pressure = Var(initialize=101325, units=pyunits.Pa)

    return m


@pytest.mark.unit
def test_relative_permitivity_constant(frame):
    relative_permittivity_constant.build_parameters(frame.params)

    assert isinstance(frame.params.relative_permittivity_liq_comp, Var)
    assert value(frame.params.relative_permittivity_liq_comp) == 101

    expr = relative_permittivity_constant.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == 101

    frame.props[1].temperature.value = 600
    assert value(expr) == 101

    assert_units_equivalent(expr, pyunits.dimensionless)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Lennard-Jones implementation of Chapman Enskog viscosity from
Properties of Gases and Liquids, 5th Ed., 9-4-1

Authors: Douglas Allan
"""

import pytest
import types

import pyomo.environ as pyo
from pyomo.environ import ConcreteModel, Block, value, units as pyunits, Var
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.ChapmanEnskog import (
    ChapmanEnskogLennardJones,
    collision_integral_neufeld_callback,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata


def construct_dummy_model(param_dict):
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    # Parameters for sulfur dioxide, from Properties of Gases and Liquids 5th ed. Appendix B.
    m.params.config.parameter_data = {
        "lennard_jones_sigma": (param_dict["sigma"], pyunits.angstrom),
        "lennard_jones_epsilon_reduced": (
            param_dict["epsilon"],
            pyunits.K,
        ),
    }

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Create variables that should exist on param block
    m.params.mw = Var(initialize=param_dict["mw"], units=pyunits.kg / pyunits.mol)
    m.params.omega = Var(initialize=param_dict["omega"], units=pyunits.dimensionless)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=273.16, units=pyunits.K)
    return m


@pytest.mark.unit
def test_visc_vap_comp_sulfur_dioxide():
    m = construct_dummy_model(
        {
            "sigma": 4.112,
            "epsilon": 335.4,
            "mw": 0.064065,
            "omega": 0.257,
        }
    )

    ChapmanEnskogLennardJones.visc_d_phase_comp.build_parameters(m.params, "Vap")

    assert isinstance(m.params.lennard_jones_sigma, Var)
    assert value(m.params.lennard_jones_sigma) == pytest.approx(4.112e-10, rel=1e-12)
    assert isinstance(m.params.lennard_jones_epsilon_reduced, Var)
    assert value(m.params.lennard_jones_epsilon_reduced) == 335.4
    assert (
        m.params.viscosity_collision_integral_callback
        is collision_integral_neufeld_callback
    )

    expr = ChapmanEnskogLennardJones.visc_d_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )
    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)

    # Pulled from Table 9.2, Properties of Gases and Liquids 5th Ed.
    T_list = [10, 100, 300, 700]  # Gas temperature in C
    visc_list = [120, 163, 246, 376]  # Experimental viscosities in millipoise
    for i in range(4):
        m.props[1].temperature.value = T_list[i] + 273.15
        assert pyo.value(expr_micropoise) == pytest.approx(visc_list[i], rel=0.03)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_visc_vap_comp_methanol():
    m = construct_dummy_model(
        {
            "sigma": 3.626,
            "epsilon": 481.8,
            "mw": 32.042e-3,
            "omega": 0.565,
        }
    )
    ChapmanEnskogLennardJones.visc_d_phase_comp.build_parameters(m.params, "Vap")

    expr = ChapmanEnskogLennardJones.visc_d_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )
    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids
    T_list = [67, 127, 277]  # Gas temperature in C
    visc_list = [112, 132, 181]  # Experimental viscosities in millipoise
    for i in range(3):
        m.props[1].temperature.value = T_list[i] + 273.15
        assert pyo.value(expr_micropoise) == pytest.approx(visc_list[i], rel=0.03)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)


@pytest.mark.unit
def test_visc_vap_comp_ethane():
    m = construct_dummy_model(
        {
            "sigma": 4.443,
            "epsilon": 215.7,
            "mw": 30.070e-3,
            "omega": 0.099,
        }
    )
    ChapmanEnskogLennardJones.visc_d_phase_comp.build_parameters(m.params, "Vap")

    expr = ChapmanEnskogLennardJones.visc_d_phase_comp.return_expression(
        m.props[1], m.params, "Vap", m.props[1].temperature
    )
    expr_micropoise = pyunits.convert(expr, pyunits.micropoise)
    # Pulled from Table 9.2, Properties of Gases and Liquids
    T_list = [47, 117, 247]  # Gas temperature in C
    visc_list = [100, 120, 156]  # Experimental viscosities in millipoise

    for i in range(3):
        m.props[1].temperature.value = T_list[i] + 273.15
        assert pyo.value(expr_micropoise) == pytest.approx(visc_list[i], rel=0.03)

    assert_units_equivalent(expr, pyunits.Pa * pyunits.s)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tray model for distillation. Can build the following:
1. Conventional tray with liq/vap inlet and liq/vap outlet
2. Feed tray with single feed inlet, liq/vap inlet, and liq/vap outlet
3. Conventional tray with side liq/vap draws (with or without feed inlet)

NOTE:
1. Does not use the IDAES control volume blocks.
2. Side vapor draw is unconventional but the model allows the user to
have a side vapor draw if required.
"""

__author__ = "Jaffer Ghouse"

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.network import Port
from pyomo.environ import (
    Var,
    Set,
    value,
    check_optimal_termination,
)

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import (
    ConfigurationError,
    InitializationError,
)
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
from idaes.models_extra.column_models.util import make_phase_split

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Tray")
class TrayData(UnitModelBlockData):
    """
    Tray unit for distillation model.
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Flash units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Flash units do not have defined volume, thus
this must be False.""",
        ),
    )
    CONFIG.declare(
        "is_feed_tray",
        ConfigValue(
            default=False,
            domain=Bool,
            description="flag to indicate feed tray.",
            doc="""indicates if this is a feed tray and constructs
corresponding ports,
**default** - False.
**Valid values:** {
**True** - feed tray,
**False** - conventional tray with no feed inlet}""",
        ),
    )
    CONFIG.declare(
        "has_liquid_side_draw",
        ConfigValue(
            default=False,
            domain=Bool,
            description="liquid side draw construction flag.",
            doc="""indicates if there is a liquid side draw from the tray,
**default** - False.
**Valid values:** {
**True** - include a liquid side draw from the tray,
**False** - exclude a liquid side draw from the tray.}""",
        ),
    )
    CONFIG.declare(
        "has_vapor_side_draw",
        ConfigValue(
            default=False,
            domain=Bool,
            description="vapor side draw construction flag.",
            doc="""indicates if there is a vapor side draw from the tray,
**default** - False.
**Valid values:** {
**True** - include a vapor side draw from the tray,
**False** - exclude a vapor side draw from the tray.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="heat duty to/from tray construction flag.",
            doc="""indicates if there is heat duty to/from the tray,
**default** - False.
**Valid values:** {
**True** - include a heat duty term,
**False** - exclude a heat duty term.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="pressure change term construction flag",
            doc="""indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="property package to use for control volume",
            doc="""property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="arguments to use for constructing property packages",
            doc="""a ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """Build the model.

        Args:
            None
        Returns:
            None
        """
        # Setup model build logger
        model_log = idaeslog.getModelLogger(self.name, tag="unit")

        # Call UnitModel.build to setup dynamics
        super(TrayData, self).build()

        # Create the inlet list to build inlet state blocks
        if self.config.is_feed_tray:
            inlet_list = ["feed", "liq", "vap"]
        else:
            inlet_list = ["liq", "vap"]

        # Create a dict to set up the inlet state blocks
        state_block_args = dict(**self.config.property_package_args)
        state_block_args["has_phase_equilibrium"] = True
        state_block_args["defined_state"] = True

        # Check for default property package
        self._get_property_package()

        for i in inlet_list:
            state_obj = self.config.property_package.build_state_block(
                self.flowsheet().time,
                doc="State block for " + i + "_inlet to tray",
                **state_block_args,
            )

            setattr(self, "properties_in_" + i, state_obj)

        # Create a dict to set up the mixed outlet state blocks
        mixed_block_args = dict(**self.config.property_package_args)
        mixed_block_args["has_phase_equilibrium"] = True
        mixed_block_args["defined_state"] = False

        self.properties_out = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="State block for mixed outlet from tray",
            **mixed_block_args,
        )

        self._add_material_balance()
        self._add_energy_balance()
        self._add_pressure_balance()

        # Get liquid and vapor phase objects from the property package
        # to be used below. Avoids repetition.
        _liquid_list = []
        _vapor_list = []
        for p in self.config.property_package.phase_list:
            pobj = self.config.property_package.get_phase(p)
            if pobj.is_vapor_phase():
                _vapor_list.append(p)
            elif pobj.is_liquid_phase():
                _liquid_list.append(p)
            else:
                _liquid_list.append(p)
                model_log.warning(
                    "A non-liquid/non-vapor phase was detected but will "
                    "be treated as a liquid."
                )

        # Create a pyomo set for indexing purposes. This set is appended to
        # model otherwise results in an abstract set.
        self._liquid_set = Set(initialize=_liquid_list)
        self._vapor_set = Set(initialize=_vapor_list)

        self._add_ports()

    def _add_material_balance(self):
        """Method to construct the mass balance equation."""

        @self.Constraint(
            self.flowsheet().time,
            self.config.property_package.component_list,
            doc="material balance",
        )
        def material_mixing_equations(b, t, j):
            if self.config.is_feed_tray:
                return 0 == sum(
                    self.properties_in_feed[t].get_material_flow_terms(p, j)
                    + self.properties_in_liq[t].get_material_flow_terms(p, j)
                    + self.properties_in_vap[t].get_material_flow_terms(p, j)
                    - self.properties_out[t].get_material_flow_terms(p, j)
                    for p in b.config.property_package.phase_list
                )
            else:
                return 0 == sum(
                    self.properties_in_liq[t].get_material_flow_terms(p, j)
                    + self.properties_in_vap[t].get_material_flow_terms(p, j)
                    - self.properties_out[t].get_material_flow_terms(p, j)
                    for p in b.config.property_package.phase_list
                )

    def _add_energy_balance(self):
        """Method to construct the energy balance equation."""

        if self.config.has_heat_transfer:
            units_meta = self.config.property_package.get_metadata()
            self.heat_duty = Var(
                self.flowsheet().time,
                initialize=0,
                doc="Heat duty for the tray",
                units=units_meta.get_derived_units("power"),
            )

        @self.Constraint(self.flowsheet().time, doc="energy balance")
        def enthalpy_mixing_equations(b, t):
            if self.config.is_feed_tray:
                if self.config.has_heat_transfer:
                    return (
                        0
                        == (
                            sum(
                                self.properties_in_feed[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                            + sum(
                                self.properties_in_liq[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                            + sum(
                                self.properties_in_vap[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                            - sum(
                                self.properties_out[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                        )
                        + self.heat_duty[t]
                    )
                else:
                    return 0 == (
                        sum(
                            self.properties_in_feed[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                        + sum(
                            self.properties_in_liq[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                        + sum(
                            self.properties_in_vap[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                        - sum(
                            self.properties_out[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                    )
            else:
                if self.config.has_heat_transfer:
                    return (
                        0
                        == (
                            sum(
                                self.properties_in_liq[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                            + sum(
                                self.properties_in_vap[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                            - sum(
                                self.properties_out[t].get_enthalpy_flow_terms(p)
                                for p in b.config.property_package.phase_list
                            )
                        )
                        + self.heat_duty[t]
                    )
                else:
                    return 0 == (
                        sum(
                            self.properties_in_liq[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                        + sum(
                            self.properties_in_vap[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                        - sum(
                            self.properties_out[t].get_enthalpy_flow_terms(p)
                            for p in b.config.property_package.phase_list
                        )
                    )

    def _add_pressure_balance(self):
        """Method to construct the pressure balance."""
        if self.config.has_pressure_change:
            units_meta = self.config.property_package.get_metadata()
            self.deltaP = Var(
                self.flowsheet().time,
                initialize=0,
                doc="Pressure drop across tray",
                units=units_meta.get_derived_units("pressure"),
            )

        @self.Constraint(self.flowsheet().time, doc="pressure balance for tray")
        def pressure_drop_equation(self, t):
            if self.config.has_pressure_change:
                return (
                    self.properties_out[t].pressure
                    == self.properties_in_liq[t].pressure - self.deltaP[t]
                )
            else:
                return (
                    self.properties_out[t].pressure
                    == self.properties_in_liq[t].pressure
                )

    def _add_ports(self):
        """Method to construct the ports for the tray."""

        # Add feed inlet port
        if self.config.is_feed_tray:
            self.add_inlet_port(name="feed", block=self.properties_in_feed)

        # Add liquid and vapor inlet ports
        self.add_inlet_port(name="liq_in", block=self.properties_in_liq)
        self.add_inlet_port(name="vap_in", block=self.properties_in_vap)

        # Add liquid outlet port
        self.liq_out = Port(noruleinit=True, doc="liquid outlet from tray")

        # Add liquid side draw port if selected
        if self.config.has_liquid_side_draw:
            self.liq_side_sf = Var(
                initialize=0.01, doc="split fraction for the liquid side draw"
            )
            self.liq_side_draw = Port(noruleinit=True, doc="liquid side draw.")
            make_phase_split(
                self,
                port=self.liq_side_draw,
                phase=self._liquid_set,
                side_sf=self.liq_side_sf,
            )

            # Populate the liquid outlet port with the remaining liquid
            # after the side draw
            make_phase_split(
                self,
                port=self.liq_out,
                phase=self._liquid_set,
                side_sf=1 - self.liq_side_sf,
            )
        else:
            # Populate the liquid outlet port when no liquid side draw
            make_phase_split(self, port=self.liq_out, phase=self._liquid_set, side_sf=1)

        # Add the vapor outlet port
        self.vap_out = Port(noruleinit=True, doc="vapor outlet from tray")

        # Add vapor side draw port if selected
        if self.config.has_vapor_side_draw:
            self.vap_side_sf = Var(
                initialize=0.01, doc="split fraction for the vapor side draw"
            )
            self.vap_side_draw = Port(noruleinit=True, doc="vapor side draw.")
            make_phase_split(
                self,
                port=self.vap_side_draw,
                phase=self._vapor_set,
                side_sf=self.vap_side_sf,
            )
            # Populate the vapor outlet port with the remaining vapor
            # after the vapor side draw
            make_phase_split(
                self,
                port=self.vap_out,
                phase=self._vapor_set,
                side_sf=1 - self.vap_side_sf,
            )
        else:
            # Populate the vapor outlet port when no vapor side draw
            make_phase_split(self, port=self.vap_out, phase=self._vapor_set, side_sf=1)

    def initialize(
        self,
        state_args_feed=None,
        state_args_liq=None,
        state_args_vap=None,
        hold_state_liq=False,
        hold_state_vap=False,
        solver=None,
        optarg=None,
        outlvl=idaeslog.NOTSET,
    ):

        # TODO:
        # 1. Initialization for dynamic mode. Currently not supported.
        # 2. Handle unfixed side split fraction vars
        # 3. Better logic to handle and fix state vars.

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        init_log.info("Begin initialization.")

        solverobj = get_solver(solver, optarg)

        if self.config.has_liquid_side_draw:
            if not self.liq_side_sf.fixed:
                raise ConfigurationError(
                    "Liquid side draw split fraction not fixed but "
                    "has_liquid_side_draw set to True."
                )

        if self.config.has_vapor_side_draw:
            if not self.vap_side_sf.fixed:
                raise ConfigurationError(
                    "Vapor side draw split fraction not fixed but "
                    "has_vapor_side_draw set to True."
                )

        # Create initial guess if not provided by using current values
        if self.config.is_feed_tray and state_args_feed is None:
            state_args_feed = {}
            state_args_liq = {}
            state_args_vap = {}
            state_dict = self.properties_in_feed[
                self.flowsheet().time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if "flow" in k:
                    if state_dict[k].is_indexed():
                        state_args_feed[k] = {}
                        state_args_liq[k] = {}
                        state_args_vap[k] = {}
                        for m in state_dict[k].keys():
                            state_args_feed[k][m] = value(state_dict[k][m])
                            state_args_liq[k][m] = value(0.1 * state_dict[k][m])
                            state_args_vap[k][m] = value(0.1 * state_dict[k][m])

                    else:
                        state_args_feed[k] = value(state_dict[k])
                        state_args_liq[k] = 0.1 * value(state_dict[k])
                        state_args_vap[k] = 0.1 * value(state_dict[k])
                else:
                    if state_dict[k].is_indexed():
                        state_args_feed[k] = {}
                        state_args_liq[k] = {}
                        state_args_vap[k] = {}
                        for m in state_dict[k].keys():
                            state_args_feed[k][m] = value(state_dict[k][m])
                            state_args_liq[k][m] = value(state_dict[k][m])
                            state_args_vap[k][m] = value(state_dict[k][m])

                    else:
                        state_args_feed[k] = value(state_dict[k])
                        state_args_liq[k] = value(state_dict[k])
                        state_args_vap[k] = value(state_dict[k])

        # Create initial guess if not provided by using current values
        if not self.config.is_feed_tray and state_args_liq is None:
            state_args_liq = {}
            state_dict = self.properties_in_liq[
                self.flowsheet().time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args_liq[k] = {}
                    for m in state_dict[k].keys():
                        state_args_liq[k][m] = value(state_dict[k][m])
                else:
                    state_args_liq[k] = value(state_dict[k])

        # Create initial guess if not provided by using current values
        if not self.config.is_feed_tray and state_args_vap is None:
            state_args_vap = {}
            state_dict = self.properties_in_vap[
                self.flowsheet().time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args_vap[k] = {}
                    for m in state_dict[k].keys():
                        state_args_vap[k][m] = value(state_dict[k][m])
                else:
                    state_args_vap[k] = value(state_dict[k])

        if self.config.is_feed_tray:
            feed_flags = self.properties_in_feed.initialize(
                outlvl=outlvl,
                solver=solver,
                optarg=optarg,
                hold_state=True,
                state_args=state_args_feed,
                state_vars_fixed=False,
            )

        liq_in_flags = self.properties_in_liq.initialize(
            outlvl=outlvl,
            solver=solver,
            optarg=optarg,
            hold_state=True,
            state_args=state_args_liq,
            state_vars_fixed=False,
        )

        vap_in_flags = self.properties_in_vap.initialize(
            outlvl=outlvl,
            solver=solver,
            optarg=optarg,
            hold_state=True,
            state_args=state_args_vap,
            state_vars_fixed=False,
        )

        # state args to initialize the mixed outlet state block
        state_args_mixed = {}

        if self.config.is_feed_tray:

            # if feed tray, initialize the mixed state block at
            # the same condition.
            state_args_mixed = state_args_feed
        else:
            # if not feed tray, initialize mixed state block at average of
            # vap/liq inlets except pressure. While this is crude, it
            # will work for most combination of state vars.
            state_dict = self.properties_in_liq[
                self.flowsheet().time.first()
            ].define_port_members()
            for k in state_dict.keys():
                if k == "pressure":
                    # Take the lowest pressure and this is the liq inlet
                    state_args_mixed[k] = value(
                        self.properties_in_liq[0].component(state_dict[k].local_name)
                    )
                elif state_dict[k].is_indexed():
                    state_args_mixed[k] = {}
                    for m in state_dict[k].keys():
                        if "flow" in k:
                            state_args_mixed[k][m] = value(
                                self.properties_in_liq[0].component(
                                    state_dict[k].local_name
                                )[m]
                            ) + value(
                                self.properties_in_vap[0].component(
                                    state_dict[k].local_name
                                )[m]
                            )

                        else:
                            state_args_mixed[k][m] = 0.5 * (
                                value(
                                    self.properties_in_liq[0].component(
                                        state_dict[k].local_name
                                    )[m]
                                )
                                + value(
                                    self.properties_in_vap[0].component(
                                        state_dict[k].local_name
                                    )[m]
                                )
                            )

                else:
                    if "flow" in k:
                        state_args_mixed[k] = value(
                            self.properties_in_liq[0].component(
                                state_dict[k].local_name
                            )
                        ) + value(
                            self.properties_in_vap[0].component(
                                state_dict[k].local_name
                            )
                        )
                    else:
                        state_args_mixed[k] = 0.5 * (
                            value(
                                self.properties_in_liq[0].component(
                                    state_dict[k].local_name
                                )
                            )
                            + value(
                                self.properties_in_vap[0].component(
                                    state_dict[k].local_name
                                )
                            )
                        )

        # Initialize the mixed outlet state block
        self.properties_out.initialize(
            outlvl=outlvl,
            solver=solver,
            optarg=optarg,
            hold_state=False,
            state_args=state_args_mixed,
            state_vars_fixed=False,
        )

        # Deactivate energy balance
        self.enthalpy_mixing_equations.deactivate()

        # Try fixing the outlet temperature if else pass
        # NOTE: if passed then there would probably be a degree of freedom
        try:
            self.properties_out[:].temperature.fix(state_args_mixed["temperature"])
        except AttributeError:
            init_log.warning(
                "Trying to fix outlet temperature "
                "during initialization but temperature attribute "
                "unavailable in the state block. Initialization "
                "proceeding with a potential degree of freedom."
            )

        # Deactivate pressure balance
        self.pressure_drop_equation.deactivate()

        # Try fixing the outlet temperature if else pass
        # NOTE: if passed then there would probably be a degree of freedom
        try:
            self.properties_out[:].pressure.fix(state_args_mixed["pressure"])
        except AttributeError:
            init_log.warning(
                "Trying to fix outlet pressure "
                "during initialization but pressure attribute "
                "unavailable in the state block. Initialization "
                "proceeding with a potential degree of freedom."
            )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self, tee=slc.tee)
        init_log.info("Mass balance solve {}.".format(idaeslog.condition(res)))

        # Activate energy balance
        self.enthalpy_mixing_equations.activate()
        try:
            self.properties_out[:].temperature.unfix()
        except AttributeError:
            pass

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self, tee=slc.tee)
        init_log.info(
            "Mass and energy balance solve {}.".format(idaeslog.condition(res))
        )

        # Activate pressure balance
        self.pressure_drop_equation.activate()
        try:
            self.properties_out[:].pressure.unfix()
        except AttributeError:
            pass

        if degrees_of_freedom(self) == 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solverobj.solve(self, tee=slc.tee)
            init_log.info(
                "Mass, energy and pressure balance solve {}.".format(
                    idaeslog.condition(res)
                )
            )
        else:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "State vars fixed but degrees of freedom "
                "for tray block is not zero during "
                "initialization."
            )

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info(
            "Initialization complete, status {}.".format(idaeslog.condition(res))
        )

        if not self.config.is_feed_tray:
            if not hold_state_vap:
                self.properties_in_vap.release_state(flags=vap_in_flags, outlvl=outlvl)
            if not hold_state_liq:
                self.properties_in_liq.release_state(flags=liq_in_flags, outlvl=outlvl)
            if hold_state_liq and hold_state_vap:
                return liq_in_flags, vap_in_flags
            elif hold_state_vap:
                return vap_in_flags
            elif hold_state_liq:
                return liq_in_flags
        else:
            self.properties_in_liq.release_state(flags=liq_in_flags, outlvl=outlvl)
            self.properties_in_vap.release_state(flags=vap_in_flags, outlvl=outlvl)
            return feed_flags
